(* En esta unidad se implementa el actor TDemandaDetallada
básicamente una demanda detallada está descripta por los valores de la demanda
para cada hora y para cada día en el horizonte de tiempo.

Los datos de la demanda son leidos de un archivo binario que tiene
al inicio un record con : el primer día de datos indicado como un TDateTime
y la fecha del día siguiente al último día de datos.

Luego en el binario hay un recor para cada día con los 24 valores correspondientes
a las potencias horarias en MW.

*)(*Se agrega la capacidad de cargar al simplex requerimientos de reserva
desde el 8/5/15 por Micho mvarela@adme.com.uy*)
unit udemandadetallada;

  {$MODE Delphi}

interface

uses
  SysUtils,
  uauxiliares,
  xmatdefs,
  uglobs,
  uDemandas, unodos, usimplex,
  ufechas,
  ucosa, ucosaConNombre, udatoshorariosdetallados,
  uconstantesSimSEE, uFuentesAleatorias, Classes, contnrs, uFichasLPD;

resourcestring
  rs_DemandaDetallada = 'Demanda Detallada';
  rs_DemandaDetalladaComienzaEn = 'Demanda detallada, comienza en';
  rs_LaSimulacionTieneFechaDeInicio = 'la simulación tiene fecha de inicio';
  rs_LaOptimizacionTieneFechaDeInicio = 'la optimización tiene fecha de inicio';
  rs_DemandaDetalladaTerminaEn = 'Demanda detallada, termina en';
  rs_LaSimulacionTieneFechaDeFin = 'la simulación tiene fecha de fin';
  rs_LaOptimizacionTieneFechaDeFin = 'la optimización tiene fecha de fin';

type

  { TDemandaDetallada }

  TDemandaDetallada = class(TDemanda)
  private
    CacheActivo: boolean;
  public
    ArchiDatos: TArchiRef;

    datos: TDatosHorariosDetallados;

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; nodo: TNodo; falla_profundidad, falla_costo: TDAOfNReal;
      ArchiDatos: string; fuente: TFuenteAleatoria; nombreBorne: string;
      SumarEnergiaHr, flg_SumarParaPostizado: boolean; icf_Fuente: TFuenteAleatoria;
      icf_NombreBorne: string; fReserva: Nreal);

    function Rec: TCosa_RecLnk; override;
    procedure BeforeRead(version, id_hilo: integer); override;
    procedure AfterRead(version, id_hilo: integer); override;

    procedure PrepararMemoria(Catalogo: TCatalogoReferencias; globs: TGlobs); override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    function PotenciaFirme: NReal; override;

    //      procedure prepararSim; override;
    procedure prepararPaso_as; override;


    procedure Free; override;
    {===Se agregan con la reserva===}
    procedure opt_cargue(s: TSimplex); override;
    {=======Micho===================}

    {$IFDEF BOSTA}
    procedure AfterInstantiation; override;
    {$ENDIF}

  end;

(* Escalones de falla por defecto
( Uruguay-invierno 2006 )
escf [pu]  0.05  0.075  0.075  0.8
cvf [USD/MWh]  140  400  1200  2000
*)

(* Deforma una curva de potencia horaria aplicando factores de crecimiento de
energía (fce) y del pico (fcp).
Por ejemplo, si el pico decrese 10% y la energía debe aumentar 5% los factores
serán fce= 1.05 y fcp= 0.9

El resultado es TRUE si todas las potencias resultantes son positivas y false
si alguna quedó negativa.
*)

procedure AlInicio;
procedure AlFinal;

implementation

(* Escalones de falla por defecto
( Uruguay-invierno 2006 )
escf [pu]  0.05  0.075  0.075  0.8
cvf [USD/MWh]  140  400  1200  2000
*)

//-----------------------------
// Métodos de TDemandaDetallada
//=============================

constructor TDemandaDetallada.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades: TFichasLPD; nodo: TNodo;
  falla_profundidad, falla_costo: TDAOfNReal; ArchiDatos: string;
  fuente: TFuenteAleatoria; nombreBorne: string;
  SumarEnergiaHr, flg_SumarParaPostizado: boolean; icf_Fuente: TFuenteAleatoria;
  icf_NombreBorne: string; fReserva: Nreal);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodo,
    falla_profundidad, falla_costo,
    fuente, nombreBorne,
    icf_Fuente, icf_NombreBorne, SumarEnergiaHr, flg_SumarParaPostizado);
  datos := nil;

  if FileExists(ArchiDatos) then
    self.ArchiDatos := TArchiRef.Create(ArchiDatos)
  else
    raise ExceptionFileNotFound.Create(ArchiDatos, 'Creando el actor ' + ClaseNombre);

  CacheActivo := False;
  Self.fReserva := fReserva;
end;

function TDemandaDetallada.Rec: TCosa_RecLnk;
begin
  Result:=inherited Rec;
  REsult.addCampoDef_archRef('ArchiDatos', archiDatos );
end;

procedure TDemandaDetallada.BeforeRead(version, id_hilo: integer);
begin
  inherited BeforeRead(version, id_hilo);
end;

procedure TDemandaDetallada.AfterRead(version, id_hilo: integer);
begin
  inherited AfterRead(version, id_hilo);
  CacheActivo := False;
  if not ArchiDatos.testearYResolver then
    raise ExceptionFileNotFound.Create(ArchiDatos.archi, 'Creando el actor ' +
      ClaseNombre);
end;

procedure TDemandaDetallada.PrepararMemoria(Catalogo: TCatalogoReferencias;
  globs: TGlobs);
begin
  inherited prepararMemoria(Catalogo, globs);

  if datos = nil then
    datos := TDatosHorariosDetallados.Create(archiDatos.archi, ucosa.lista_caminos);

  if globs.EstadoDeLaSala <> CES_SIN_PREPARAR then
  begin
    if (globs.fechaIniSim.AsDt < datos.fechaPrimerDia) then
      raise Exception.Create(rs_DemandaDetalladaComienzaEn + ': ' +
        DateTimeToStr(datos.FechaPrimerDia) + ' ' +
        rs_LaSimulacionTieneFechaDeInicio + ': ' + globs.fechaIniSim.AsStr);
    if (globs.fechaIniOpt.AsDt < datos.fechaPrimerDia) then
      raise Exception.Create(rs_DemandaDetalladaComienzaEn + ': ' +
        DateTimeToStr(datos.FechaPrimerDia) + ' ' +
        rs_LaOptimizacionTieneFechaDeInicio + ': ' + globs.fechaIniOpt.AsStr);

    if globs.fechaFinSim.asDt >= (datos.fechaUltimodia + 1) then
      raise Exception.Create(rs_DemandaDetalladaTerminaEn + ': ' +
        DateTimeToStr(datos.FechaUltimoDia + 23.0 / 24.0) + ' ' +
        rs_LaSimulacionTieneFechaDeFin + ': ' + globs.fechaFinSim.AsStr);
    if globs.fechaFinOpt.asDt >= (datos.fechaUltimodia + 1) then
      raise Exception.Create(rs_DemandaDetalladaTerminaEn + ': ' +
        DateTimeToStr(datos.FechaUltimoDia + 23.0 / 24.0) + ' ' +
        rs_LaOptimizacionTieneFechaDeFin + ': ' + globs.fechaFinOpt.AsStr);
  end;
end;

function TDemandaDetallada.InfoAd_: string;
begin
  Result := ArchiDatos.archi;
end;

class function TDemandaDetallada.DescClase: string;
begin
  Result := rs_DemandaDetallada;
end;

function TDemandaDetallada.PotenciaFirme: NReal;
begin
  Result := paUnidades.nUnidades_Operativas[0] * vmax(self.PPa);
end;


procedure TDemandaDetallada.prepararPaso_as;
begin
  if globs.SalaMinutal then
    PHoraria_SinAfectar[0] := datos.interpol(globs.FechaInicioDelPaso.AsDt)
  else
    datos.ReadBuff_horario( PHoraria_SinAfectar, globs.FechaInicioDelpaso.AsDt);
  inherited prepararPaso_as;
end;


procedure TDemandaDetallada.Free;
begin
  if ArchiDatos <> nil then
  begin
    ArchiDatos.Free;
    ArchiDatos := nil;
  end;

  if datos <> nil then
    datos.Free;
  inherited Free;
end;


{===Lo podría agregar si quiero agregar un despacho de falla de reserva==}{
procedure TDemandaDetallada.opt_nvers(var ivar, ivae, ires: integer);
begin
end;
 }

procedure TDemandaDetallada.opt_cargue(s: TSimplex);

var
  iposte, iescalon: integer;
  ibaseres: integer;
begin
  if NMaquinasDisponibles = 0 then
    exit;
  inherited opt_cargue(s);
  if (globs.flg_ReservaRotante) then
  begin
    {$IFDEF SPXCONLOG}
    spx_NombrarVariables(s);
    {$ENDIF}
    ibaseres := nodo.ires;
    // Carga el término constante de las restricciones de demanda de Reserva
    for iposte := 0 to High(PPa) do
      s.acum_e(ibaseres + Length(PPA) + iposte, s.nc, -PPa[iposte] * fReserva);
  end;
end;

{$IFDEF BOSTA}
procedure TDemandaDetallada.AfterInstantiation;
begin
  inherited AfterInstantiation;

  CacheActivo := False;
  if not ArchiDatos.testearYResolver then
    raise ExceptionFileNotFound.Create(ArchiDatos.archi, 'Creando el actor ' +
      ClaseNombre);
end;
{$ENDIF}

procedure AlInicio;
begin
  registrarClaseDeCosa(TDemandaDetallada.ClassName, TDemandaDetallada);
end;

procedure AlFinal;
begin
end;

end.
